program paleom87;

{Forward and inverse transformation for the polyconical projection used
in the 'Atlas of Lithological-Paleogeographical Maps of the World -
Mesozoic and Cenozoic of continents and oceans' (Ronov et al., 1989).

written by: Michael Schulz
            GEOMAR
            Research Center for Marine Geosciences
            Wischhofstr. 1-3
            D-24148 Kiel

            e-mail: mschulz@geomar.uni-kiel.d400.de


Required data format for ASCII input-files:

              x1 y1
              x2 y2
              x3 y3      1st line segment
              .  .
              .  .
              xn yn
              -9999
              x1 y1
              x2 y2
              x3 y3      2nd line segment
              .  .
              .  .
              xn yn
              -9999

A Flag (-9999) is used to indicate the end of each line segment. This flag
is also required if the input data-file contains only one line segment!

The output-file (ASCII or ATLAS format) is written to the same path as the
input-file. The extension is changed to '.DAT'. If '.DAT' was the extension
of the input-file then the output extension will be '.NEW'.

Data format of the ASCII output-files:

              lon_1 lat_1
              lon_2 lat_2
              lon_3 lat_3       1st line segment
                .     .
                .     .
              lon_n lat_n
              -9999
              lon_1 lat_1
              lon_2 lat_2
              lon_3 lat_3       2nd line segment
                .     .
                .     .
              lon_n lat_n
              -9999


The Program can be run in batchmode. In this case wildcards can be used for
the filename but not for the extension:

Syntax: PALEOM87 Filename [-Fasc]

        Filename: Complete path of the input file
        F: Output-file type
           asc -> ASCII-file, default
           atl -> ATLAS-file

        e.g.: PALEOM87 c:\data\myfile.xy -Fatl

}

{$M 65520,0,655360}  {the Newton routines require a large stack}
{$G+,N+,E-}
{$IFDEF Make}
    {$R-,S-,D-,L-,I-,V-,B-}
{$ENDIF}

uses crt, dos;

const
     stdlon: double = 40.0;            {central meridian = 40E}
     m     : double = 1.0/(48.0e+6);   {numerical value of the map scale}

var
   x, y, lon, lat: double;
   i             : byte;
   key           : char;
   auto, asc     : boolean;
   fname         : pathstr;


{-------------------------------------------------------------------------}
{calculate a^b, with b being a positive integer by multiplying a
 b-times with itself}
function ipow(a: double; b: word): double; external;
{$l ipow287d.obj}


{-------------------------------------------------------------------------}
{lat/lon-coordinate to x/y point in [mm]}
procedure c2xy(lat, lon: double; var x: double; var y: double);
var
  lon2, lon3, lat2, lat4: double;

begin
    lon:= lon - stdlon * pi/180.0;
    lon2:= lon * lon;
    lon3:= lon2 * lon;
    lat2:= lat * lat;
    lat4:= lat2 * lat2;

    y:= 6366927995.0 * lat;
    y:= y + 234641741.8 * lat * lon2;
    y:= y - 94229585.2 * ipow(lat, 3) * lon2;
    y:= y + 23131445.1 * ipow(lat, 5) * lon2;
    y:= y * m;

    x:= 5371771740.6 * lon;
    x:= x - 1487638498.8 * lat2 * lon;
    x:= x + 372496080.0 * lat4 * lon;
    x:= x - 94756618.4 * ipow(lat, 6) * lon;
    x:= x - 6180779.6 * lat2 *lon3;
    x:= x + 1236741.6 * lat4 * lon3;
    x:= x * m;
end;


{-------------------------------------------------------------------------}
{$I xy2cnewt.pas}

procedure xy2c(xin, yin: double; var lat: double; var lon: double);
const
    eps = 1.0E-6;
var
  NofEQ, EofM: boolean;

begin

{---determine quadrant}
    if xin >= 0.0 then EofM:= true;
    if xin < 0.0 then EofM:= false;
    if yin > 0.0 then NofEQ:= true;
    if yin < 0.0 then NofEQ:= false;
    if ((abs(xin) <= eps) and (abs(yin) <= eps)) then begin
       lat:= 0.0;
       lon:= stdlon;
       exit;
    end;
    xy2cnewt (xin, yin, lat, lon, stdlon, m);
    if (EofM = true)  then lon:= lon;
    if (EofM = true) and (lon > 180.0)   then lon:= lon - 360.0;
    if (EofM = false) and (lon <= 260.0) then lon:= -1.0 * (lon - 2.0*stdlon);
    if (EofM = false) and (lon > 260.0)  then lon:= 360.0 - (lon - 2.0*stdlon);
    if NofEQ = true then lat:= lat;
    if NofEQ = false then lat:= -1.0 * lat;

{---make sure that points located on the central meridian or the equator
    have the desired coordinates}
    if (xin = 0.0) then lon:= stdlon;
    if (yin = 0.0) then lat:= 0.0;
end;


{-------------------------------------------------------------------------}
procedure coor2xy;
begin
    repeat
        clrscr;
        gotoxy(5,5);
        write('Enter longitude: ');
        readln(lon);
        gotoxy(5,6);
        write('Enter latitude : ');
        readln(lat);
        lat:= lat * pi/180;
        lon:= lon * pi/180;
        c2xy(lat, lon, x, y);
        gotoxy(5,9);
        writeln('x = ',x:7:2,' mm relative to central meridian');
        gotoxy(5,10);
        writeln('y = ',y:7:2,' mm relative to Equator');
        gotoxy(5,13);
        write('....again [y/n]');
        repeat
             key:= readkey;
        until (key = 'y') or (key = 'n');
    until key = 'n';
end;


{-------------------------------------------------------------------------}
procedure xy2coor;
begin
    repeat
        clrscr;
        gotoxy(5,5);
        write('Enter x in [mm] relative to central meridian (E -> +): ');
        readln(x);
        gotoxy(5,6);
        write('Enter y in [mm] relative to Equator (N -> +)         : ');
        readln(y);
        xy2c(x, y, lat, lon);
        gotoxy(5,9);
        writeln('lon = ',lon:7:2,'');
        gotoxy(5,10);
        writeln('lat = ',lat:7:2,'');
        gotoxy(5,13);
        write('....again [y/n]');
        repeat
             key:= readkey;
        until (key = 'y') or (key = 'n');
    until key = 'n';
end;


{-------------------------------------------------------------------------}
procedure xy2dat(auto: boolean; fname: string);

const
    maxdim = 8000;
type
    rlvector  = array [1..maxdim] of real;
    rlvectptr = ^rlvector;
    exvector  = array [1..maxdim] of double;
    exvectptr = ^exvector;
var
   dx, dy    : rlvectptr;                    {read real values from file;
                                             double type results in garbage}
   dlat, dlon: exvectptr;
   i, npoint,
   cnt, nline: integer;
   key       : char;
   outname   : PathStr;
   dir       : DirStr;
   name      : NameStr;
   ext       : ExtStr;
   inunit,
   outunit   : text;
label 1, 2;

begin
    new(dx); new(dy); new(dlat); new(dlon);

{---open input-file}
    if auto = false then begin
       clrscr;
       gotoxy(5,5);
       write('Enter complete Filename with X/Y-Data: ');
       readln(fname);
    end;
    assign(inunit, fname);
    {$I-}
    reset(inunit);
    {$I+}
    if ioresult <> 0 then begin
       writeln; writeln;
       writeln('    ***************');
       writeln('    * I/O-Error ! *');
       writeln('    ***************');
       writeln;
       write('    Try again ? [y/n]');
       repeat key:= readkey;
       until (key = 'y') or (key = 'n');
       if key = 'y' then begin
          dispose(dx); dispose(dy); dispose(dlat); dispose(dlon);
          xy2dat(auto, fname);

{---------since program returns to this line after an error and recursive calling of xy2dat}
          exit;
       end
       else {key = 'n'} begin
          dispose(dx); dispose(dy); dispose(dlat); dispose(dlon);
          exit;
       end;
    end;

{---open output-file}
    repeat
        clrscr;
        gotoxy(5,5);
        fsplit(fname, dir, name, ext);
        if ((ext <> '.DAT') and (ext <> '.dat')) then ext:= '.DAT'
        else ext:= '.NEW';
        outname:= dir + name + ext;
        write(' Saving transformed data to ', outname);
        assign(outunit, outname);
        {$I-}
        rewrite(outunit);
        {$I+}
        if ioresult <> 0 then begin
           writeln; writeln;
           writeln('    ***************');
           writeln('    * I/O-Error ! *');
           writeln('    ***************');
           writeln;
           write('    Press any key to return to the main menu...');
           repeat until keypressed;
            dispose(dx); dispose(dy); dispose(dlat); dispose(dlon);
           exit;
        end;
    until ioresult = 0;
    if asc = false then begin
       writeln(outunit,'FORMAT');
       writeln(outunit,'(10F8.2)');
       writeln(outunit,'READ');
    end;

{---read line segment}
    nline:= 1;
    repeat
         i:= 1;
         repeat
              read(inunit, dx^[i]);
              if dx^[i] = -9999 then begin
                 readln(inunit);                  {jump to the next line}
                 goto 1;
              end;
              if eof(inunit) then goto 2;
              readln (inunit, dy^[i]);
              inc (i,1);
1:       until (dx^[i] = -9999);

{--------delete end of line flag in order to prepare for the next loop}
         dx^[i]:= 0;
         npoint:= i-1;

{--------transform line segment}
         clrscr;
         gotoxy(5,5);
         case asc of
            true : write('Saving converted data from ',fname,' to ASCII Data File');
            false: write('Saving converted data from ',fname,' to ATLAS Data File');
         end;
         gotoxy(5,7);
         write('Converting Line-Segment:             Point: ');
         for i:= 1 to npoint do begin
             gotoxy(31,7);
             write(nline:3);
             gotoxy(50,7);
             write(i:4);
             xy2c(dx^[i], dy^[i], dlat^[i], dlon^[i]);
         end;

{--------save lat/lon data into output-file}
{--------ASCII File}
         case asc of
            true: begin
                      for i:= 1 to npoint do begin
                         write(outunit, dlon^[i]:8:2,'  ');
                         writeln(outunit, dlat^[i]:8:2);
                      end;
                      writeln(outunit, '   -9999');
                  end;

{-----------ATLAS File}
            false: begin
                       writeln(outunit,npoint+1:4);
                       write(outunit,' 1000.0     0.0 ');
                       for i:= 1 to 4 do begin
                          write(outunit, dlat^[i]:8:2);
                          write(outunit, dlon^[i]:8:2);
                       end;
                       writeln(outunit,'');
                       cnt:= 0;
                       for i:= 5 to npoint do begin
                          inc(cnt, 1);
                          write(outunit, dlat^[i]:8:2);
                          write(outunit, dlon^[i]:8:2);
                          if cnt = 5 then begin
                             cnt:= 0;
                             writeln(outunit,'');
                          end;
                       end;
                       if cnt <> 0 then writeln(outunit,'');
                   end;
         end;
         flush(outunit);
         inc(nline, 1);
2:  until eof(inunit);
    close(inunit);
    if asc = false then begin
       writeln(outunit,'   0');
       writeln(outunit,'');
    end;
    flush(outunit);
    close(outunit);
    dispose(dx); dispose(dy); dispose(dlat); dispose(dlon);
end;


{------------------------------------------------------------------------}
procedure bathelp;
begin
    clrscr;
    gotoxy(1,5);
    writeln('     Syntax: PALEOM87 Filename [-Fasc]');
    writeln;
    writeln('       Filename: Complete path of the input file');
    writeln('       F: Output-file type');
    writeln('          ASC -> ASCII-file, default');
    writeln('          ATL -> ATLAS-file');
    writeln;
    writeln('       e.g.: PALEOM87 c:\data\myfile.xy -Fatl');
    writeln;writeln;
    writeln('             The output-file is written to the same path as the');
    writeln('             input-file. The extension is changed to ''.DAT''. If');
    writeln('             ''.DAT'' was the extension of the input-file then the');
    writeln('             extension will be ''.NEW''.');
    halt;
end;


{------------------------------------------------------------------------}
procedure baterror;
begin
    writeln; writeln;
    writeln('    Error in command line paramter!');
    writeln;
    writeln('    Press any key for syntax help.');
    repeat until keypressed;
    bathelp;
end;


{------------------------------------------------------------------------}
{read file type from command line}
procedure getftype (pos: byte);
var
  s      : string;
  code, i: integer;

begin
    s:= copy (paramstr(pos), 1, 2);
    if (s <> '-F') and (s <> '-f') then baterror;

{---get rid of commandline switch}
    s:= copy (paramstr(pos), 3, length(paramstr(pos)));

{---convert string to upper-case}
    for i:= 1 to length(s) do s[i]:= upcase(s[i]);
    if (s <> 'ASC') and (s <> 'ATL') then baterror;
    if s = 'ASC'
       then asc:= true
       else asc:= false;
end;


{------------------------------------------------------------------------}
procedure batchmode;

var
   dirinfo  : searchrec;
   mname    : array[1..50] of PathStr;
   dir      : DirStr;
   name     : NameStr;
   ext      : ExtStr;
   i, nfiles: byte;
   mult     : boolean;
   s1, s2   : string;

begin
    if (paramcount > 2) or (paramstr(1) = '?') or (paramstr(1) = '-?') or
       (paramstr(1) = 'h') or (paramstr(1) = '-h') then bathelp;
    auto:= true;
    fname:= fexpand(paramstr(1));
    fsplit(fname, dir, name, ext);
    if name = '*' then mult:= true
    else mult:= false;
    if ext = '.*' then begin
        writeln; writeln;
        writeln('    ***************');
        writeln('    * I/O-Error ! *');
        writeln('    ***************');
        writeln;
        writeln('    Extension needs to be specified!');
        writeln;
        writeln('    Press any key for syntax help.');
        repeat until keypressed;
        bathelp;
    end;

{---transform multiple files}
    i:= 1;
    if mult = true then begin
       findfirst(fname, archive, dirinfo);
       while doserror = 0 do begin
             mname[i]:= dir + dirinfo.name;
             inc(i, 1);
             findnext(dirinfo);
       end;
       nfiles:= i - 1;
    end

{---one file only}
    else begin
       mname[1]:= fname;
       nfiles:= 1;
    end;

{---determine switch setting and assign default values}
{---no switch set}
    if (paramcount = 1) then begin
        asc:= true;
    end;

{---1 switch set in command line}
    if (paramcount = 2) then begin
        s1:= copy (paramstr(2), 2, 1);
        if (upcase (s1[1]) <> 'F') then baterror
        else getftype(2);
    end;
    for i:= 1 to nfiles do xy2dat(auto, mname[i]);
end;


{-------------------------------------------------------------------------}
begin
    if test8087 < 2 then begin
       writeln; writeln;
       writeln(' ERROR: This Version requires at least a 80287 numeric coprocessor');
       writeln('        Try PALEOMAP.EXE...');
       halt;
    end;
    if paramcount >= 1 then batchmode;
    if paramcount < 1 then begin
       repeat
            auto:= false;
            clrscr;
            gotoxy(1,3);
            writeln('              *****************************************************');
            writeln('              *  Atlas of Lithological-Paleogeographical Maps of  *');
            writeln('              *  the World - Mesozoic and Cenozoic of Continents  *');
            writeln('              *  and Oceans  (Ronov, Khain & Balukhovskiy, 1989)  *');
            writeln('              *                                                   *');
            writeln('              *        Forward and Inverse Transformation         *');
            writeln('              *                                                   *');
            writeln('              *               Version 1.4 ( 286/87)              *');
            writeln('              *****************************************************');
            gotoxy(1,15);
            writeln('       1. lon/lat -> x/y');
            writeln;
            writeln('       2. x/y -> lon/lat');
            writeln;
            writeln('       3. X/Y-File -> ASCII Data-File');
            writeln;
            writeln('       4. X/Y-File -> ATLAS Data-File');
            writeln;
            write  ('       5. Quit');
            repeat
                 key:= readkey;
            until ((key='1')or(key='2')or(key='3')or(key='4')or(key='5'));
            case key of
               '1': coor2xy;
               '2': xy2coor;
               '3': begin
                        asc:= true;
                        xy2dat(auto,'');
                    end;
               '4': begin
                        asc:= false;
                        xy2dat(auto,'');
                    end;
            end;
        until key = '5';
        clrscr;
    end;
end.
